"""Bindings generated for interface datapath by rpclib"""

from __future__ import print_function, division

import argparse
import json
import logging
import sys
import traceback

get_str = str
if sys.version_info[0] > 2:
    long = int
    unicode = str
    str = bytes

from xapi import *


class Datapath_server_dispatcher(object):
    """
    Xapi will call the functions here on VM start / shutdown /
    suspend / resume / migrate. Every function is idempotent. Every
    function takes a domain parameter which allows the implementation
    to track how many domains are currently using the volume.

    Volumes must be attached via the following sequence of calls:

    1. [open url persistent] must be called first and is used to declare
    that the writes to the disks must either be persisted or not.
    [open] is not an exclusive operation - a disk may be opened on
    more than once host at once. The call returns `unit` or an
    error.

    2. [attach url domain] is then called. The `domain` parameter is
    advisory. Note that this call is currently only ever called once.
    In the future the call may be made multiple times with different
    [domain] parameters if the disk is attached to multiple domains.
    The return value from this call is the information required to
    attach the disk to a VM. This call is again, not exclusive. The
    volume may be attached to more than one host concurrently.

    3. [activate url domain] is called to activate the datapath. This
    must be called before the volume is to be used by the VM, and
    it is acceptible for this to be an exclusive operation, such that
    it is an error for a volume to be activated on more than one host
    simultaneously.

    """

    def __init__(self, impl):
        """impl is a proxy object whose methods contain the implementation"""
        self._impl = impl
        self._dispatcher_dict = {
            "Datapath.open": self.open,
            "Datapath.attach": self.attach,
            "Datapath.activate": self.activate,
            "Datapath.activate_readonly": self.activate_readonly,
            "Datapath.import_activate": self.import_activate,
            "Datapath.deactivate": self.deactivate,
            "Datapath.detach": self.detach,
            "Datapath.close": self.close
        }

    def open(self, args):
        """type-check inputs, call implementation, type-check outputs and return"""
        if not isinstance(args, dict):
            raise UnmarshalException('arguments', 'dict', repr(args))
        if "dbg" not in args:
            raise UnmarshalException('argument missing', 'dbg', '')
        dbg = args["dbg"]
        if not isinstance(dbg, (str, unicode)):
            raise TypeError("string", repr(dbg))
        if "uri" not in args:
            raise UnmarshalException('argument missing', 'uri', '')
        uri = args["uri"]
        if not isinstance(uri, (str, unicode)):
            raise TypeError("string", repr(uri))
        if "persistent" not in args:
            raise UnmarshalException('argument missing', 'persistent', '')
        persistent = args["persistent"]
        if not isinstance(persistent, bool):
            raise TypeError("bool", repr(persistent))
        results = self._impl.open(dbg, uri, persistent)
        return results

    def attach(self, args):
        """type-check inputs, call implementation, type-check outputs and return"""
        if not isinstance(args, dict):
            raise UnmarshalException('arguments', 'dict', repr(args))
        if "dbg" not in args:
            raise UnmarshalException('argument missing', 'dbg', '')
        dbg = args["dbg"]
        if not isinstance(dbg, (str, unicode)):
            raise TypeError("string", repr(dbg))
        if "uri" not in args:
            raise UnmarshalException('argument missing', 'uri', '')
        uri = args["uri"]
        if not isinstance(uri, (str, unicode)):
            raise TypeError("string", repr(uri))
        if "domain" not in args:
            raise UnmarshalException('argument missing', 'domain', '')
        domain = args["domain"]
        if not isinstance(domain, (str, unicode)):
            raise TypeError("string", repr(domain))
        results = self._impl.attach(dbg, uri, domain)
        if not isinstance(results['implementations'], list):
            raise TypeError("| XenDisk ({ params: string; extra: (string * string) list; backend_type: string; }) (** This value can be used for ring connection. *) | BlockDevice ({ path: string; }) (** This value can be used for Domain0 block device access. *) | File ({ path: string; }) (**  *) | Nbd ({ uri: string; }) (**  *) list", repr(results['implementations']))
        for tmp_47 in results['implementations']:
            if tmp_47[0] not in ["XenDisk", "BlockDevice", "File", "Nbd"]:
                raise TypeError("| XenDisk ({ params: string; extra: (string * string) list; backend_type: string; }) (** This value can be used for ring connection. *) | BlockDevice ({ path: string; }) (** This value can be used for Domain0 block device access. *) | File ({ path: string; }) (**  *) | Nbd ({ uri: string; }) (**  *)", repr(tmp_47))
            if tmp_47[0] == 'XenDisk':
                if not isinstance(tmp_47[1]['params'], (str, unicode)):
                    raise TypeError("string", repr(tmp_47[1]['params']))
                if not isinstance(tmp_47[1]['extra'], dict):
                    raise TypeError("(string * string) list", repr(tmp_47[1]['extra']))
                for tmp_48, tmp_49 in tmp_47[1]['extra'].items():
                    if not isinstance(tmp_48, (str, unicode)):
                        raise TypeError("string", repr(tmp_48))
                    if not isinstance(tmp_49, (str, unicode)):
                        raise TypeError("string", repr(tmp_49))
                if not isinstance(tmp_47[1]['backend_type'], (str, unicode)):
                    raise TypeError("string", repr(tmp_47[1]['backend_type']))
            elif tmp_47[0] == 'BlockDevice':
                if not isinstance(tmp_47[1]['path'], (str, unicode)):
                    raise TypeError("string", repr(tmp_47[1]['path']))
            elif tmp_47[0] == 'File':
                if not isinstance(tmp_47[1]['path'], (str, unicode)):
                    raise TypeError("string", repr(tmp_47[1]['path']))
            elif tmp_47[0] == 'Nbd':
                if not isinstance(tmp_47[1]['uri'], (str, unicode)):
                    raise TypeError("string", repr(tmp_47[1]['uri']))
        return results

    def activate(self, args):
        """type-check inputs, call implementation, type-check outputs and return"""
        if not isinstance(args, dict):
            raise UnmarshalException('arguments', 'dict', repr(args))
        if "dbg" not in args:
            raise UnmarshalException('argument missing', 'dbg', '')
        dbg = args["dbg"]
        if not isinstance(dbg, (str, unicode)):
            raise TypeError("string", repr(dbg))
        if "uri" not in args:
            raise UnmarshalException('argument missing', 'uri', '')
        uri = args["uri"]
        if not isinstance(uri, (str, unicode)):
            raise TypeError("string", repr(uri))
        if "domain" not in args:
            raise UnmarshalException('argument missing', 'domain', '')
        domain = args["domain"]
        if not isinstance(domain, (str, unicode)):
            raise TypeError("string", repr(domain))
        results = self._impl.activate(dbg, uri, domain)
        return results

    def activate_readonly(self, args):
        """type-check inputs, call implementation, type-check outputs and return"""
        if not isinstance(args, dict):
            raise UnmarshalException('arguments', 'dict', repr(args))
        if "dbg" not in args:
            raise UnmarshalException('argument missing', 'dbg', '')
        dbg = args["dbg"]
        if not isinstance(dbg, (str, unicode)):
            raise TypeError("string", repr(dbg))
        if "uri" not in args:
            raise UnmarshalException('argument missing', 'uri', '')
        uri = args["uri"]
        if not isinstance(uri, (str, unicode)):
            raise TypeError("string", repr(uri))
        if "domain" not in args:
            raise UnmarshalException('argument missing', 'domain', '')
        domain = args["domain"]
        if not isinstance(domain, (str, unicode)):
            raise TypeError("string", repr(domain))
        results = self._impl.activate_readonly(dbg, uri, domain)
        return results

    def import_activate(self, args):
        """type-check inputs, call implementation, type-check outputs and return"""
        if not isinstance(args, dict):
            raise UnmarshalException('arguments', 'dict', repr(args))
        if "dbg" not in args:
            raise UnmarshalException('argument missing', 'dbg', '')
        dbg = args["dbg"]
        if not isinstance(dbg, (str, unicode)):
            raise TypeError("string", repr(dbg))
        if "uri" not in args:
            raise UnmarshalException('argument missing', 'uri', '')
        uri = args["uri"]
        if not isinstance(uri, (str, unicode)):
            raise TypeError("string", repr(uri))
        if "domain" not in args:
            raise UnmarshalException('argument missing', 'domain', '')
        domain = args["domain"]
        if not isinstance(domain, (str, unicode)):
            raise TypeError("string", repr(domain))
        results = self._impl.import_activate(dbg, uri, domain)
        if not isinstance(results, (str, unicode)):
            raise TypeError("string", repr(results))
        return results

    def deactivate(self, args):
        """type-check inputs, call implementation, type-check outputs and return"""
        if not isinstance(args, dict):
            raise UnmarshalException('arguments', 'dict', repr(args))
        if "dbg" not in args:
            raise UnmarshalException('argument missing', 'dbg', '')
        dbg = args["dbg"]
        if not isinstance(dbg, (str, unicode)):
            raise TypeError("string", repr(dbg))
        if "uri" not in args:
            raise UnmarshalException('argument missing', 'uri', '')
        uri = args["uri"]
        if not isinstance(uri, (str, unicode)):
            raise TypeError("string", repr(uri))
        if "domain" not in args:
            raise UnmarshalException('argument missing', 'domain', '')
        domain = args["domain"]
        if not isinstance(domain, (str, unicode)):
            raise TypeError("string", repr(domain))
        results = self._impl.deactivate(dbg, uri, domain)
        return results

    def detach(self, args):
        """type-check inputs, call implementation, type-check outputs and return"""
        if not isinstance(args, dict):
            raise UnmarshalException('arguments', 'dict', repr(args))
        if "dbg" not in args:
            raise UnmarshalException('argument missing', 'dbg', '')
        dbg = args["dbg"]
        if not isinstance(dbg, (str, unicode)):
            raise TypeError("string", repr(dbg))
        if "uri" not in args:
            raise UnmarshalException('argument missing', 'uri', '')
        uri = args["uri"]
        if not isinstance(uri, (str, unicode)):
            raise TypeError("string", repr(uri))
        if "domain" not in args:
            raise UnmarshalException('argument missing', 'domain', '')
        domain = args["domain"]
        if not isinstance(domain, (str, unicode)):
            raise TypeError("string", repr(domain))
        results = self._impl.detach(dbg, uri, domain)
        return results

    def close(self, args):
        """type-check inputs, call implementation, type-check outputs and return"""
        if not isinstance(args, dict):
            raise UnmarshalException('arguments', 'dict', repr(args))
        if "dbg" not in args:
            raise UnmarshalException('argument missing', 'dbg', '')
        dbg = args["dbg"]
        if not isinstance(dbg, (str, unicode)):
            raise TypeError("string", repr(dbg))
        if "uri" not in args:
            raise UnmarshalException('argument missing', 'uri', '')
        uri = args["uri"]
        if not isinstance(uri, (str, unicode)):
            raise TypeError("string", repr(uri))
        results = self._impl.close(dbg, uri)
        return results

    def _dispatch(self, method, params):
        """type check inputs, call implementation, type check outputs and return"""
        args = params[0]
        return success(self._dispatcher_dict[method](args))


class Datapath_skeleton(object):
    """
    Xapi will call the functions here on VM start / shutdown /
    suspend / resume / migrate. Every function is idempotent. Every
    function takes a domain parameter which allows the implementation
    to track how many domains are currently using the volume.

    Volumes must be attached via the following sequence of calls:

    1. [open url persistent] must be called first and is used to declare
    that the writes to the disks must either be persisted or not.
    [open] is not an exclusive operation - a disk may be opened on
    more than once host at once. The call returns `unit` or an
    error.

    2. [attach url domain] is then called. The `domain` parameter is
    advisory. Note that this call is currently only ever called once.
    In the future the call may be made multiple times with different
    [domain] parameters if the disk is attached to multiple domains.
    The return value from this call is the information required to
    attach the disk to a VM. This call is again, not exclusive. The
    volume may be attached to more than one host concurrently.

    3. [activate url domain] is called to activate the datapath. This
    must be called before the volume is to be used by the VM, and
    it is acceptible for this to be an exclusive operation, such that
    it is an error for a volume to be activated on more than one host
    simultaneously.

    """

    def __init__(self):
        pass

    def open(self, dbg, uri, persistent):
        """
        [open uri persistent] is called before a disk is attached to a VM.
        If persistent is true then care should be taken to persist all writes
        to the disk. If persistent is false then the implementation should
        configure a temporary location for writes so they can be thrown away
        on [close].
        """
        raise Unimplemented("Datapath.open")

    def attach(self, dbg, uri, domain):
        """
        [attach uri domain] prepares a connection between the storage named by
        [uri] and the Xen domain with id [domain]. The return value is the
        information needed by the Xen toolstack to setup the shared-memory
        blkfront protocol. Note that the same volume may be simultaneously
        attached to multiple hosts for example over a migrate. If an
        implementation needs to perform an explicit handover, then it should
        implement [activate] and [deactivate]. This function is idempotent.
        """
        raise Unimplemented("Datapath.attach")

    def activate(self, dbg, uri, domain):
        """
        [activate uri domain] is called just before a VM needs to read or write
        its disk. This is an opportunity for an implementation which needs to
        perform an explicit volume handover to do it. This function is called
        in the migration downtime window so delays here will be noticeable to
        users and should be minimised. This function is idempotent.
        """
        raise Unimplemented("Datapath.activate")

    def activate_readonly(self, dbg, uri, domain):
        """
        [activate_readonly uri domain] is called just before a VM, or the
        control domain, needs to read a volume. A single volume may be
        activated readonly multiple times, including on multiple independent
        hosts. It is not permitted for a volume to be activated both readonly
        and read-write concurrently. Implementations shall declare the
        VDI_ACTIVATE_READONLY feature for this method to be supported. Once a
        volume is activated readonly it is required that all readonly
        activations are deactivated before any read-write activation is
        attempted. This function is idempotent and in all other respects
        is interchangeable with activate.
        """
        raise Unimplemented("Datapath.activate_readonly")

    def import_activate(self, dbg, uri, domain):
        """
        [import_activate uri domain] prepares a connection to the
        storage named by [uri] for use by inbound import mirroring,
        the [domain] parameter identifies which domain to connect to,
        most likely 0 or a custom storage domain. The return value is a
        path to a UNIX domain socket to which an open file descriptor
        may be passed, by SCM_RIGHTS. This, in turn, will become
        the server end of a Network Block Device (NBD) connection
        using, new-fixed protocol. Implementations shall declare the
        VDI_MIRROR_IN feature for this method to be supported. It is
        expected that activate will have been previously called so that
        there is an active datapath.
        """
        raise Unimplemented("Datapath.import_activate")

    def deactivate(self, dbg, uri, domain):
        """
        [deactivate uri domain] is called as soon as a VM has finished reading
        or writing its disk. This is an opportunity for an implementation which
        needs to perform an explicit volume handover to do it. This function is
        called in the migration downtime window so delays here will be
        noticeable to users and should be minimised. This function is idempotent.
        """
        raise Unimplemented("Datapath.deactivate")

    def detach(self, dbg, uri, domain):
        """
        [detach uri domain] is called sometime after a VM has finished reading
        or writing its disk. This is an opportunity to clean up any resources
        associated with the disk. This function is called outside the migration
        downtime window so can be slow without affecting users. This function is
        idempotent. This function should never fail. If an implementation is
        unable to perform some cleanup right away then it should queue the
        action internally. Any error result represents a bug in the
        implementation.
        """
        raise Unimplemented("Datapath.detach")

    def close(self, dbg, uri):
        """
        [close uri] is called after a disk is detached and a VM shutdown. This
        is an opportunity to throw away writes if the disk is not persistent.
        """
        raise Unimplemented("Datapath.close")


class Datapath_test(object):
    """
    Xapi will call the functions here on VM start / shutdown /
    suspend / resume / migrate. Every function is idempotent. Every
    function takes a domain parameter which allows the implementation
    to track how many domains are currently using the volume.

    Volumes must be attached via the following sequence of calls:

    1. [open url persistent] must be called first and is used to declare
    that the writes to the disks must either be persisted or not.
    [open] is not an exclusive operation - a disk may be opened on
    more than once host at once. The call returns `unit` or an
    error.

    2. [attach url domain] is then called. The `domain` parameter is
    advisory. Note that this call is currently only ever called once.
    In the future the call may be made multiple times with different
    [domain] parameters if the disk is attached to multiple domains.
    The return value from this call is the information required to
    attach the disk to a VM. This call is again, not exclusive. The
    volume may be attached to more than one host concurrently.

    3. [activate url domain] is called to activate the datapath. This
    must be called before the volume is to be used by the VM, and
    it is acceptible for this to be an exclusive operation, such that
    it is an error for a volume to be activated on more than one host
    simultaneously.

    """

    def __init__(self):
        pass

    def open(self, dbg, uri, persistent):
        """
        [open uri persistent] is called before a disk is attached to a VM.
        If persistent is true then care should be taken to persist all writes
        to the disk. If persistent is false then the implementation should
        configure a temporary location for writes so they can be thrown away
        on [close].
        """

    def attach(self, dbg, uri, domain):
        """
        [attach uri domain] prepares a connection between the storage named by
        [uri] and the Xen domain with id [domain]. The return value is the
        information needed by the Xen toolstack to setup the shared-memory
        blkfront protocol. Note that the same volume may be simultaneously
        attached to multiple hosts for example over a migrate. If an
        implementation needs to perform an explicit handover, then it should
        implement [activate] and [deactivate]. This function is idempotent.
        """
        return {"implementations": [None]}

    def activate(self, dbg, uri, domain):
        """
        [activate uri domain] is called just before a VM needs to read or write
        its disk. This is an opportunity for an implementation which needs to
        perform an explicit volume handover to do it. This function is called
        in the migration downtime window so delays here will be noticeable to
        users and should be minimised. This function is idempotent.
        """

    def activate_readonly(self, dbg, uri, domain):
        """
        [activate_readonly uri domain] is called just before a VM, or the
        control domain, needs to read a volume. A single volume may be
        activated readonly multiple times, including on multiple independent
        hosts. It is not permitted for a volume to be activated both readonly
        and read-write concurrently. Implementations shall declare the
        VDI_ACTIVATE_READONLY feature for this method to be supported. Once a
        volume is activated readonly it is required that all readonly
        activations are deactivated before any read-write activation is
        attempted. This function is idempotent and in all other respects
        is interchangeable with activate.
        """

    def import_activate(self, dbg, uri, domain):
        """
        [import_activate uri domain] prepares a connection to the
        storage named by [uri] for use by inbound import mirroring,
        the [domain] parameter identifies which domain to connect to,
        most likely 0 or a custom storage domain. The return value is a
        path to a UNIX domain socket to which an open file descriptor
        may be passed, by SCM_RIGHTS. This, in turn, will become
        the server end of a Network Block Device (NBD) connection
        using, new-fixed protocol. Implementations shall declare the
        VDI_MIRROR_IN feature for this method to be supported. It is
        expected that activate will have been previously called so that
        there is an active datapath.
        """
        return "string"

    def deactivate(self, dbg, uri, domain):
        """
        [deactivate uri domain] is called as soon as a VM has finished reading
        or writing its disk. This is an opportunity for an implementation which
        needs to perform an explicit volume handover to do it. This function is
        called in the migration downtime window so delays here will be
        noticeable to users and should be minimised. This function is idempotent.
        """

    def detach(self, dbg, uri, domain):
        """
        [detach uri domain] is called sometime after a VM has finished reading
        or writing its disk. This is an opportunity to clean up any resources
        associated with the disk. This function is called outside the migration
        downtime window so can be slow without affecting users. This function is
        idempotent. This function should never fail. If an implementation is
        unable to perform some cleanup right away then it should queue the
        action internally. Any error result represents a bug in the
        implementation.
        """

    def close(self, dbg, uri):
        """
        [close uri] is called after a disk is detached and a VM shutdown. This
        is an opportunity to throw away writes if the disk is not persistent.
        """


class Datapath_commandline(object):
    """Parse command-line arguments and call an implementation."""

    def __init__(self, impl):
        self.impl = impl
        self.dispatcher = Datapath_server_dispatcher(self.impl)

    def _parse_open(self):
        """
        [open uri persistent] is called before a disk is attached to a VM.
        If persistent is true then care should be taken to persist all writes
        to the disk. If persistent is false then the implementation should
        configure a temporary location for writes so they can be thrown away
        on [close].
        """
        # in --json mode we don't have any other arguments
        if ('--json' in sys.argv or '-j' in sys.argv):
            jsondict = json.loads(sys.stdin.readline(),)
            jsondict['json'] = True
            return jsondict
        parser = argparse.ArgumentParser(description='[open uri persistent] is called before a disk is attached to a VM. If persistent is true then care should be taken to persist all writes to the disk. If persistent is false then the implementation should configure a temporary location for writes so they can be thrown away on [close].')
        parser.add_argument('-j', '--json', action='store_const', const=True, default=False, help='Read json from stdin, print json to stdout', required=False)
        parser.add_argument('dbg', action='store', help='Debug context from the caller')
        parser.add_argument('uri', action='store', help='A URI which represents how to access the volume disk data.')
        parser.add_argument('persistent', action='store', help='True means the disk data is persistent and should be preserved when the datapath is closed i.e. when a VM is shutdown or rebooted. False means the data should be thrown away when the VM is shutdown or rebooted.', type=lambda x: json.loads(x.lower()))
        return vars(parser.parse_args())

    def _parse_attach(self):
        """
        [attach uri domain] prepares a connection between the storage named by
        [uri] and the Xen domain with id [domain]. The return value is the
        information needed by the Xen toolstack to setup the shared-memory
        blkfront protocol. Note that the same volume may be simultaneously
        attached to multiple hosts for example over a migrate. If an
        implementation needs to perform an explicit handover, then it should
        implement [activate] and [deactivate]. This function is idempotent.
        """
        # in --json mode we don't have any other arguments
        if ('--json' in sys.argv or '-j' in sys.argv):
            jsondict = json.loads(sys.stdin.readline(),)
            jsondict['json'] = True
            return jsondict
        parser = argparse.ArgumentParser(description='[attach uri domain] prepares a connection between the storage named by [uri] and the Xen domain with id [domain]. The return value is the information needed by the Xen toolstack to setup the shared-memory blkfront protocol. Note that the same volume may be simultaneously attached to multiple hosts for example over a migrate. If an implementation needs to perform an explicit handover, then it should implement [activate] and [deactivate]. This function is idempotent.')
        parser.add_argument('-j', '--json', action='store_const', const=True, default=False, help='Read json from stdin, print json to stdout', required=False)
        parser.add_argument('dbg', action='store', help='Debug context from the caller')
        parser.add_argument('uri', action='store', help='A URI which represents how to access the volume disk data.')
        parser.add_argument('domain', action='store', help='An opaque string which represents the Xen domain.')
        return vars(parser.parse_args())

    def _parse_activate(self):
        """
        [activate uri domain] is called just before a VM needs to read or write
        its disk. This is an opportunity for an implementation which needs to
        perform an explicit volume handover to do it. This function is called
        in the migration downtime window so delays here will be noticeable to
        users and should be minimised. This function is idempotent.
        """
        # in --json mode we don't have any other arguments
        if ('--json' in sys.argv or '-j' in sys.argv):
            jsondict = json.loads(sys.stdin.readline(),)
            jsondict['json'] = True
            return jsondict
        parser = argparse.ArgumentParser(description='[activate uri domain] is called just before a VM needs to read or write its disk. This is an opportunity for an implementation which needs to perform an explicit volume handover to do it. This function is called in the migration downtime window so delays here will be noticeable to users and should be minimised. This function is idempotent.')
        parser.add_argument('-j', '--json', action='store_const', const=True, default=False, help='Read json from stdin, print json to stdout', required=False)
        parser.add_argument('dbg', action='store', help='Debug context from the caller')
        parser.add_argument('uri', action='store', help='A URI which represents how to access the volume disk data.')
        parser.add_argument('domain', action='store', help='An opaque string which represents the Xen domain.')
        return vars(parser.parse_args())

    def _parse_activate_readonly(self):
        """
        [activate_readonly uri domain] is called just before a VM, or the
        control domain, needs to read a volume. A single volume may be
        activated readonly multiple times, including on multiple independent
        hosts. It is not permitted for a volume to be activated both readonly
        and read-write concurrently. Implementations shall declare the
        VDI_ACTIVATE_READONLY feature for this method to be supported. Once a
        volume is activated readonly it is required that all readonly
        activations are deactivated before any read-write activation is
        attempted. This function is idempotent and in all other respects
        is interchangeable with activate.
        """
        # in --json mode we don't have any other arguments
        if ('--json' in sys.argv or '-j' in sys.argv):
            jsondict = json.loads(sys.stdin.readline(),)
            jsondict['json'] = True
            return jsondict
        parser = argparse.ArgumentParser(description='[activate_readonly uri domain] is called just before a VM, or the  control domain, needs to read a volume. A single volume may be  activated readonly multiple times, including on multiple independent  hosts. It is not permitted for a volume to be activated both readonly  and read-write concurrently. Implementations shall declare the  VDI_ACTIVATE_READONLY feature for this method to be supported. Once a  volume is activated readonly it is required that all readonly  activations are deactivated before any read-write activation is  attempted. This function is idempotent and in all other respects  is interchangeable with activate.')
        parser.add_argument('-j', '--json', action='store_const', const=True, default=False, help='Read json from stdin, print json to stdout', required=False)
        parser.add_argument('dbg', action='store', help='Debug context from the caller')
        parser.add_argument('uri', action='store', help='A URI which represents how to access the volume disk data.')
        parser.add_argument('domain', action='store', help='An opaque string which represents the Xen domain.')
        return vars(parser.parse_args())

    def _parse_import_activate(self):
        """
        [import_activate uri domain] prepares a connection to the
        storage named by [uri] for use by inbound import mirroring,
        the [domain] parameter identifies which domain to connect to,
        most likely 0 or a custom storage domain. The return value is a
        path to a UNIX domain socket to which an open file descriptor
        may be passed, by SCM_RIGHTS. This, in turn, will become
        the server end of a Network Block Device (NBD) connection
        using, new-fixed protocol. Implementations shall declare the
        VDI_MIRROR_IN feature for this method to be supported. It is
        expected that activate will have been previously called so that
        there is an active datapath.
        """
        # in --json mode we don't have any other arguments
        if ('--json' in sys.argv or '-j' in sys.argv):
            jsondict = json.loads(sys.stdin.readline(),)
            jsondict['json'] = True
            return jsondict
        parser = argparse.ArgumentParser(description='[import_activate uri domain] prepares a connection to the   storage named by [uri] for use by inbound import mirroring,  the [domain] parameter identifies which domain to connect to,  most likely 0 or a custom storage domain. The return value is a  path to a UNIX domain socket to which an open file descriptor  may be passed, by SCM_RIGHTS. This, in turn, will become  the server end of a Network Block Device (NBD) connection  using, new-fixed protocol. Implementations shall declare the  VDI_MIRROR_IN feature for this method to be supported. It is  expected that activate will have been previously called so that  there is an active datapath.')
        parser.add_argument('-j', '--json', action='store_const', const=True, default=False, help='Read json from stdin, print json to stdout', required=False)
        parser.add_argument('dbg', action='store', help='Debug context from the caller')
        parser.add_argument('uri', action='store', help='A URI which represents how to access the volume disk data.')
        parser.add_argument('domain', action='store', help='An opaque string which represents the Xen domain.')
        return vars(parser.parse_args())

    def _parse_deactivate(self):
        """
        [deactivate uri domain] is called as soon as a VM has finished reading
        or writing its disk. This is an opportunity for an implementation which
        needs to perform an explicit volume handover to do it. This function is
        called in the migration downtime window so delays here will be
        noticeable to users and should be minimised. This function is idempotent.
        """
        # in --json mode we don't have any other arguments
        if ('--json' in sys.argv or '-j' in sys.argv):
            jsondict = json.loads(sys.stdin.readline(),)
            jsondict['json'] = True
            return jsondict
        parser = argparse.ArgumentParser(description='[deactivate uri domain] is called as soon as a VM has finished reading or writing its disk. This is an opportunity for an implementation which needs to perform an explicit volume handover to do it. This function is called in the migration downtime window so delays here will be noticeable to users and should be minimised. This function is idempotent.')
        parser.add_argument('-j', '--json', action='store_const', const=True, default=False, help='Read json from stdin, print json to stdout', required=False)
        parser.add_argument('dbg', action='store', help='Debug context from the caller')
        parser.add_argument('uri', action='store', help='A URI which represents how to access the volume disk data.')
        parser.add_argument('domain', action='store', help='An opaque string which represents the Xen domain.')
        return vars(parser.parse_args())

    def _parse_detach(self):
        """
        [detach uri domain] is called sometime after a VM has finished reading
        or writing its disk. This is an opportunity to clean up any resources
        associated with the disk. This function is called outside the migration
        downtime window so can be slow without affecting users. This function is
        idempotent. This function should never fail. If an implementation is
        unable to perform some cleanup right away then it should queue the
        action internally. Any error result represents a bug in the
        implementation.
        """
        # in --json mode we don't have any other arguments
        if ('--json' in sys.argv or '-j' in sys.argv):
            jsondict = json.loads(sys.stdin.readline(),)
            jsondict['json'] = True
            return jsondict
        parser = argparse.ArgumentParser(description='[detach uri domain] is called sometime after a VM has finished reading or writing its disk. This is an opportunity to clean up any resources associated with the disk. This function is called outside the migration downtime window so can be slow without affecting users. This function is idempotent. This function should never fail. If an implementation is unable to perform some cleanup right away then it should queue the action internally. Any error result represents a bug in the implementation.')
        parser.add_argument('-j', '--json', action='store_const', const=True, default=False, help='Read json from stdin, print json to stdout', required=False)
        parser.add_argument('dbg', action='store', help='Debug context from the caller')
        parser.add_argument('uri', action='store', help='A URI which represents how to access the volume disk data.')
        parser.add_argument('domain', action='store', help='An opaque string which represents the Xen domain.')
        return vars(parser.parse_args())

    def _parse_close(self):
        """
        [close uri] is called after a disk is detached and a VM shutdown. This
        is an opportunity to throw away writes if the disk is not persistent.
        """
        # in --json mode we don't have any other arguments
        if ('--json' in sys.argv or '-j' in sys.argv):
            jsondict = json.loads(sys.stdin.readline(),)
            jsondict['json'] = True
            return jsondict
        parser = argparse.ArgumentParser(description='[close uri] is called after a disk is detached and a VM shutdown. This is an opportunity to throw away writes if the disk is not persistent.')
        parser.add_argument('-j', '--json', action='store_const', const=True, default=False, help='Read json from stdin, print json to stdout', required=False)
        parser.add_argument('dbg', action='store', help='Debug context from the caller')
        parser.add_argument('uri', action='store', help='A URI which represents how to access the volume disk data.')
        return vars(parser.parse_args())

    def open(self):
        """
        [open uri persistent] is called before a disk is attached to a VM.
        If persistent is true then care should be taken to persist all writes
        to the disk. If persistent is false then the implementation should
        configure a temporary location for writes so they can be thrown away
        on [close].
        """
        use_json = False
        try:
            request = self._parse_open()
            use_json = 'json' in request and request['json']
            results = self.dispatcher.open(request)
            print(json.dumps(results))
        except Exception as exn:
            if use_json:
                handle_exception(exn)
            else:
                traceback.print_exc()
                raise exn

    def attach(self):
        """
        [attach uri domain] prepares a connection between the storage named by
        [uri] and the Xen domain with id [domain]. The return value is the
        information needed by the Xen toolstack to setup the shared-memory
        blkfront protocol. Note that the same volume may be simultaneously
        attached to multiple hosts for example over a migrate. If an
        implementation needs to perform an explicit handover, then it should
        implement [activate] and [deactivate]. This function is idempotent.
        """
        use_json = False
        try:
            request = self._parse_attach()
            use_json = 'json' in request and request['json']
            results = self.dispatcher.attach(request)
            print(json.dumps(results))
        except Exception as exn:
            if use_json:
                handle_exception(exn)
            else:
                traceback.print_exc()
                raise exn

    def activate(self):
        """
        [activate uri domain] is called just before a VM needs to read or write
        its disk. This is an opportunity for an implementation which needs to
        perform an explicit volume handover to do it. This function is called
        in the migration downtime window so delays here will be noticeable to
        users and should be minimised. This function is idempotent.
        """
        use_json = False
        try:
            request = self._parse_activate()
            use_json = 'json' in request and request['json']
            results = self.dispatcher.activate(request)
            print(json.dumps(results))
        except Exception as exn:
            if use_json:
                handle_exception(exn)
            else:
                traceback.print_exc()
                raise exn

    def activate_readonly(self):
        """
        [activate_readonly uri domain] is called just before a VM, or the
        control domain, needs to read a volume. A single volume may be
        activated readonly multiple times, including on multiple independent
        hosts. It is not permitted for a volume to be activated both readonly
        and read-write concurrently. Implementations shall declare the
        VDI_ACTIVATE_READONLY feature for this method to be supported. Once a
        volume is activated readonly it is required that all readonly
        activations are deactivated before any read-write activation is
        attempted. This function is idempotent and in all other respects
        is interchangeable with activate.
        """
        use_json = False
        try:
            request = self._parse_activate_readonly()
            use_json = 'json' in request and request['json']
            results = self.dispatcher.activate_readonly(request)
            print(json.dumps(results))
        except Exception as exn:
            if use_json:
                handle_exception(exn)
            else:
                traceback.print_exc()
                raise exn

    def import_activate(self):
        """
        [import_activate uri domain] prepares a connection to the
        storage named by [uri] for use by inbound import mirroring,
        the [domain] parameter identifies which domain to connect to,
        most likely 0 or a custom storage domain. The return value is a
        path to a UNIX domain socket to which an open file descriptor
        may be passed, by SCM_RIGHTS. This, in turn, will become
        the server end of a Network Block Device (NBD) connection
        using, new-fixed protocol. Implementations shall declare the
        VDI_MIRROR_IN feature for this method to be supported. It is
        expected that activate will have been previously called so that
        there is an active datapath.
        """
        use_json = False
        try:
            request = self._parse_import_activate()
            use_json = 'json' in request and request['json']
            results = self.dispatcher.import_activate(request)
            print(json.dumps(results))
        except Exception as exn:
            if use_json:
                handle_exception(exn)
            else:
                traceback.print_exc()
                raise exn

    def deactivate(self):
        """
        [deactivate uri domain] is called as soon as a VM has finished reading
        or writing its disk. This is an opportunity for an implementation which
        needs to perform an explicit volume handover to do it. This function is
        called in the migration downtime window so delays here will be
        noticeable to users and should be minimised. This function is idempotent.
        """
        use_json = False
        try:
            request = self._parse_deactivate()
            use_json = 'json' in request and request['json']
            results = self.dispatcher.deactivate(request)
            print(json.dumps(results))
        except Exception as exn:
            if use_json:
                handle_exception(exn)
            else:
                traceback.print_exc()
                raise exn

    def detach(self):
        """
        [detach uri domain] is called sometime after a VM has finished reading
        or writing its disk. This is an opportunity to clean up any resources
        associated with the disk. This function is called outside the migration
        downtime window so can be slow without affecting users. This function is
        idempotent. This function should never fail. If an implementation is
        unable to perform some cleanup right away then it should queue the
        action internally. Any error result represents a bug in the
        implementation.
        """
        use_json = False
        try:
            request = self._parse_detach()
            use_json = 'json' in request and request['json']
            results = self.dispatcher.detach(request)
            print(json.dumps(results))
        except Exception as exn:
            if use_json:
                handle_exception(exn)
            else:
                traceback.print_exc()
                raise exn

    def close(self):
        """
        [close uri] is called after a disk is detached and a VM shutdown. This
        is an opportunity to throw away writes if the disk is not persistent.
        """
        use_json = False
        try:
            request = self._parse_close()
            use_json = 'json' in request and request['json']
            results = self.dispatcher.close(request)
            print(json.dumps(results))
        except Exception as exn:
            if use_json:
                handle_exception(exn)
            else:
                traceback.print_exc()
                raise exn


class Data_server_dispatcher(object):
    """
    This interface is used for long-running data operations such as
    copying the contents of volumes or mirroring volumes to remote
    destinations.

    These operations are asynchronous and rely on the Tasks API to
    report results and errors.

    To mirror a VDI a sequence of these API calls is required:

    1. Create a destination VDI using the Volume API on the destination
    SR. This must be the same size as the source. To minimize copying
    the destination VDI may be cloned from one that has been previously
    copied, as long as a disk from which the copy was made is still
    present on the source (even as a metadata-only disk)

    2. Arrange for the destination disk to be accessible on the source
    host by suitable URL. This may be `nbd`, `iscsi`, `nfs` or
    other URL.

    3. Start mirroring all new writes to the destination disk via the
    `Data.mirror` API call.

    4. Find the list of blocks to copy via the CBT API call. Note that if
    the destination volume has not been 'prezeroed' then all of the
    blocks must be copied to the destination.

    5. Start the background copy of the disk via a call to `DATA.copy`,
    passing in the list of blocks to copy. The plugin must ensure that
    the copy does not conflict with the mirror operation - ie., all
    writes from the mirror operation must not be overwritten by writes
    of old data from the copy operation.

    6. The progress of the copy operation may be queried via the `Data.stat`
    call.

    7. Once the copy operation has succesfully completed the destination
    disk will be a perfect mirror of the source.


    """

    def __init__(self, impl):
        """impl is a proxy object whose methods contain the implementation"""
        self._impl = impl
        self._dispatcher_dict = {
            "Data.copy": self.copy,
            "Data.mirror": self.mirror,
            "Data.stat": self.stat,
            "Data.cancel": self.cancel,
            "Data.destroy": self.destroy,
            "Data.ls": self.ls
        }

    def copy(self, args):
        """type-check inputs, call implementation, type-check outputs and return"""
        if not isinstance(args, dict):
            raise UnmarshalException('arguments', 'dict', repr(args))
        if "dbg" not in args:
            raise UnmarshalException('argument missing', 'dbg', '')
        dbg = args["dbg"]
        if not isinstance(dbg, (str, unicode)):
            raise TypeError("string", repr(dbg))
        if "uri" not in args:
            raise UnmarshalException('argument missing', 'uri', '')
        uri = args["uri"]
        if not isinstance(uri, (str, unicode)):
            raise TypeError("string", repr(uri))
        if "domain" not in args:
            raise UnmarshalException('argument missing', 'domain', '')
        domain = args["domain"]
        if not isinstance(domain, (str, unicode)):
            raise TypeError("string", repr(domain))
        if "remote" not in args:
            raise UnmarshalException('argument missing', 'remote', '')
        remote = args["remote"]
        if not isinstance(remote, (str, unicode)):
            raise TypeError("string", repr(remote))
        if "blocklist" not in args:
            raise UnmarshalException('argument missing', 'blocklist', '')
        blocklist = args["blocklist"]
        if not isinstance(blocklist['blocksize'], (int, long)):
            raise TypeError("int", repr(blocklist['blocksize']))
        if not isinstance(blocklist['ranges'], list):
            raise TypeError("(int64 * int64) list", repr(blocklist['ranges']))
        for tmp_50 in blocklist['ranges']:
            if not (isinstance(tmp_50, tuple) and len(tmp_50) == 2):
                raise TypeError("(int64 * int64)", repr(tmp_50))
            v1, v2 = tmp_50
            if not isinstance(v1, (int, long)):
                raise TypeError("int64", repr(v1))
            if not isinstance(v2, (int, long)):
                raise TypeError("int64", repr(v2))
        results = self._impl.copy(dbg, uri, domain, remote, blocklist)
        if results[0] not in ["CopyV1", "MirrorV1"]:
            raise TypeError("| CopyV1 (string) (** CopyV1 (key) represents an on-going copy operation with the unique [key]. *) | MirrorV1 (string) (** MirrorV1 (key) represents an on-going mirror operation with the unique [key]. *)", repr(results))
        if results[0] == 'CopyV1':
            if not isinstance(results[1], (str, unicode)):
                raise TypeError("string", repr(results[1]))
        elif results[0] == 'MirrorV1':
            if not isinstance(results[1], (str, unicode)):
                raise TypeError("string", repr(results[1]))
        return results

    def mirror(self, args):
        """type-check inputs, call implementation, type-check outputs and return"""
        if not isinstance(args, dict):
            raise UnmarshalException('arguments', 'dict', repr(args))
        if "dbg" not in args:
            raise UnmarshalException('argument missing', 'dbg', '')
        dbg = args["dbg"]
        if not isinstance(dbg, (str, unicode)):
            raise TypeError("string", repr(dbg))
        if "uri" not in args:
            raise UnmarshalException('argument missing', 'uri', '')
        uri = args["uri"]
        if not isinstance(uri, (str, unicode)):
            raise TypeError("string", repr(uri))
        if "domain" not in args:
            raise UnmarshalException('argument missing', 'domain', '')
        domain = args["domain"]
        if not isinstance(domain, (str, unicode)):
            raise TypeError("string", repr(domain))
        if "remote" not in args:
            raise UnmarshalException('argument missing', 'remote', '')
        remote = args["remote"]
        if not isinstance(remote, (str, unicode)):
            raise TypeError("string", repr(remote))
        results = self._impl.mirror(dbg, uri, domain, remote)
        if results[0] not in ["CopyV1", "MirrorV1"]:
            raise TypeError("| CopyV1 (string) (** CopyV1 (key) represents an on-going copy operation with the unique [key]. *) | MirrorV1 (string) (** MirrorV1 (key) represents an on-going mirror operation with the unique [key]. *)", repr(results))
        if results[0] == 'CopyV1':
            if not isinstance(results[1], (str, unicode)):
                raise TypeError("string", repr(results[1]))
        elif results[0] == 'MirrorV1':
            if not isinstance(results[1], (str, unicode)):
                raise TypeError("string", repr(results[1]))
        return results

    def stat(self, args):
        """type-check inputs, call implementation, type-check outputs and return"""
        if not isinstance(args, dict):
            raise UnmarshalException('arguments', 'dict', repr(args))
        if "dbg" not in args:
            raise UnmarshalException('argument missing', 'dbg', '')
        dbg = args["dbg"]
        if not isinstance(dbg, (str, unicode)):
            raise TypeError("string", repr(dbg))
        if "operation" not in args:
            raise UnmarshalException('argument missing', 'operation', '')
        operation = args["operation"]
        if operation[0] not in ["CopyV1", "MirrorV1"]:
            raise TypeError("| CopyV1 (string) (** CopyV1 (key) represents an on-going copy operation with the unique [key]. *) | MirrorV1 (string) (** MirrorV1 (key) represents an on-going mirror operation with the unique [key]. *)", repr(operation))
        if operation[0] == 'CopyV1':
            if not isinstance(operation[1], (str, unicode)):
                raise TypeError("string", repr(operation[1]))
        elif operation[0] == 'MirrorV1':
            if not isinstance(operation[1], (str, unicode)):
                raise TypeError("string", repr(operation[1]))
        results = self._impl.stat(dbg, operation)
        if not isinstance(results['failed'], bool):
            raise TypeError("bool", repr(results['failed']))
        if not isinstance(results['complete'], bool):
            raise TypeError("bool", repr(results['complete']))
        if 'progress' in results:
            if not isinstance(results['progress'], float):
                raise TypeError("float", repr(results['progress']))
        return results

    def cancel(self, args):
        """type-check inputs, call implementation, type-check outputs and return"""
        if not isinstance(args, dict):
            raise UnmarshalException('arguments', 'dict', repr(args))
        if "dbg" not in args:
            raise UnmarshalException('argument missing', 'dbg', '')
        dbg = args["dbg"]
        if not isinstance(dbg, (str, unicode)):
            raise TypeError("string", repr(dbg))
        if "operation" not in args:
            raise UnmarshalException('argument missing', 'operation', '')
        operation = args["operation"]
        if operation[0] not in ["CopyV1", "MirrorV1"]:
            raise TypeError("| CopyV1 (string) (** CopyV1 (key) represents an on-going copy operation with the unique [key]. *) | MirrorV1 (string) (** MirrorV1 (key) represents an on-going mirror operation with the unique [key]. *)", repr(operation))
        if operation[0] == 'CopyV1':
            if not isinstance(operation[1], (str, unicode)):
                raise TypeError("string", repr(operation[1]))
        elif operation[0] == 'MirrorV1':
            if not isinstance(operation[1], (str, unicode)):
                raise TypeError("string", repr(operation[1]))
        results = self._impl.cancel(dbg, operation)
        return results

    def destroy(self, args):
        """type-check inputs, call implementation, type-check outputs and return"""
        if not isinstance(args, dict):
            raise UnmarshalException('arguments', 'dict', repr(args))
        if "dbg" not in args:
            raise UnmarshalException('argument missing', 'dbg', '')
        dbg = args["dbg"]
        if not isinstance(dbg, (str, unicode)):
            raise TypeError("string", repr(dbg))
        if "operation" not in args:
            raise UnmarshalException('argument missing', 'operation', '')
        operation = args["operation"]
        if operation[0] not in ["CopyV1", "MirrorV1"]:
            raise TypeError("| CopyV1 (string) (** CopyV1 (key) represents an on-going copy operation with the unique [key]. *) | MirrorV1 (string) (** MirrorV1 (key) represents an on-going mirror operation with the unique [key]. *)", repr(operation))
        if operation[0] == 'CopyV1':
            if not isinstance(operation[1], (str, unicode)):
                raise TypeError("string", repr(operation[1]))
        elif operation[0] == 'MirrorV1':
            if not isinstance(operation[1], (str, unicode)):
                raise TypeError("string", repr(operation[1]))
        results = self._impl.destroy(dbg, operation)
        return results

    def ls(self, args):
        """type-check inputs, call implementation, type-check outputs and return"""
        if not isinstance(args, dict):
            raise UnmarshalException('arguments', 'dict', repr(args))
        if "dbg" not in args:
            raise UnmarshalException('argument missing', 'dbg', '')
        dbg = args["dbg"]
        if not isinstance(dbg, (str, unicode)):
            raise TypeError("string", repr(dbg))
        results = self._impl.ls(dbg)
        if not isinstance(results, list):
            raise TypeError("| CopyV1 (string) (** CopyV1 (key) represents an on-going copy operation with the unique [key]. *) | MirrorV1 (string) (** MirrorV1 (key) represents an on-going mirror operation with the unique [key]. *) list", repr(results))
        for tmp_51 in results:
            if tmp_51[0] not in ["CopyV1", "MirrorV1"]:
                raise TypeError("| CopyV1 (string) (** CopyV1 (key) represents an on-going copy operation with the unique [key]. *) | MirrorV1 (string) (** MirrorV1 (key) represents an on-going mirror operation with the unique [key]. *)", repr(tmp_51))
            if tmp_51[0] == 'CopyV1':
                if not isinstance(tmp_51[1], (str, unicode)):
                    raise TypeError("string", repr(tmp_51[1]))
            elif tmp_51[0] == 'MirrorV1':
                if not isinstance(tmp_51[1], (str, unicode)):
                    raise TypeError("string", repr(tmp_51[1]))
        return results

    def _dispatch(self, method, params):
        """type check inputs, call implementation, type check outputs and return"""
        args = params[0]
        return success(self._dispatcher_dict[method](args))


class Data_skeleton(object):
    """
    This interface is used for long-running data operations such as
    copying the contents of volumes or mirroring volumes to remote
    destinations.

    These operations are asynchronous and rely on the Tasks API to
    report results and errors.

    To mirror a VDI a sequence of these API calls is required:

    1. Create a destination VDI using the Volume API on the destination
    SR. This must be the same size as the source. To minimize copying
    the destination VDI may be cloned from one that has been previously
    copied, as long as a disk from which the copy was made is still
    present on the source (even as a metadata-only disk)

    2. Arrange for the destination disk to be accessible on the source
    host by suitable URL. This may be `nbd`, `iscsi`, `nfs` or
    other URL.

    3. Start mirroring all new writes to the destination disk via the
    `Data.mirror` API call.

    4. Find the list of blocks to copy via the CBT API call. Note that if
    the destination volume has not been 'prezeroed' then all of the
    blocks must be copied to the destination.

    5. Start the background copy of the disk via a call to `DATA.copy`,
    passing in the list of blocks to copy. The plugin must ensure that
    the copy does not conflict with the mirror operation - ie., all
    writes from the mirror operation must not be overwritten by writes
    of old data from the copy operation.

    6. The progress of the copy operation may be queried via the `Data.stat`
    call.

    7. Once the copy operation has succesfully completed the destination
    disk will be a perfect mirror of the source.


    """

    def __init__(self):
        pass

    def copy(self, dbg, uri, domain, remote, blocklist):
        """
        [copy uri domain remotes blocks] copies [blocks] from the local disk
        to a remote URI. This may be called as part of a Volume Mirroring
        operation, and hence may need to cooperate with whatever process is
        currently mirroring writes to ensure data integrity is maintained.
        The [remote] parameter is a remotely accessible URI, for example,
        `nbd://root:pass@foo.com/path/to/disk` that must contain all necessary
        authentication tokens
        """
        raise Unimplemented("Data.copy")

    def mirror(self, dbg, uri, domain, remote):
        """
        [mirror uri domain remote] starts mirroring new writes to the volume
        to a remote URI (usually NBD). This is called as part of a volume
        mirroring process
        """
        raise Unimplemented("Data.mirror")

    def stat(self, dbg, operation):
        """
        [stat operation] returns the current status of [operation]. For a
        copy operation, this will contain progress information.
        """
        raise Unimplemented("Data.stat")

    def cancel(self, dbg, operation):
        """
        [cancel operation] cancels a long-running operation. Note that the
        call may return before the operation has finished.
        """
        raise Unimplemented("Data.cancel")

    def destroy(self, dbg, operation):
        """
        [destroy operation] destroys the information about a long-running
        operation. This should fail when run against an operation that is
        still in progress.
        """
        raise Unimplemented("Data.destroy")

    def ls(self, dbg):
        """
        [ls] returns a list of all current operations
        """
        raise Unimplemented("Data.ls")


class Data_test(object):
    """
    This interface is used for long-running data operations such as
    copying the contents of volumes or mirroring volumes to remote
    destinations.

    These operations are asynchronous and rely on the Tasks API to
    report results and errors.

    To mirror a VDI a sequence of these API calls is required:

    1. Create a destination VDI using the Volume API on the destination
    SR. This must be the same size as the source. To minimize copying
    the destination VDI may be cloned from one that has been previously
    copied, as long as a disk from which the copy was made is still
    present on the source (even as a metadata-only disk)

    2. Arrange for the destination disk to be accessible on the source
    host by suitable URL. This may be `nbd`, `iscsi`, `nfs` or
    other URL.

    3. Start mirroring all new writes to the destination disk via the
    `Data.mirror` API call.

    4. Find the list of blocks to copy via the CBT API call. Note that if
    the destination volume has not been 'prezeroed' then all of the
    blocks must be copied to the destination.

    5. Start the background copy of the disk via a call to `DATA.copy`,
    passing in the list of blocks to copy. The plugin must ensure that
    the copy does not conflict with the mirror operation - ie., all
    writes from the mirror operation must not be overwritten by writes
    of old data from the copy operation.

    6. The progress of the copy operation may be queried via the `Data.stat`
    call.

    7. Once the copy operation has succesfully completed the destination
    disk will be a perfect mirror of the source.


    """

    def __init__(self):
        pass

    def copy(self, dbg, uri, domain, remote, blocklist):
        """
        [copy uri domain remotes blocks] copies [blocks] from the local disk
        to a remote URI. This may be called as part of a Volume Mirroring
        operation, and hence may need to cooperate with whatever process is
        currently mirroring writes to ensure data integrity is maintained.
        The [remote] parameter is a remotely accessible URI, for example,
        `nbd://root:pass@foo.com/path/to/disk` that must contain all necessary
        authentication tokens
        """
        return None

    def mirror(self, dbg, uri, domain, remote):
        """
        [mirror uri domain remote] starts mirroring new writes to the volume
        to a remote URI (usually NBD). This is called as part of a volume
        mirroring process
        """
        return None

    def stat(self, dbg, operation):
        """
        [stat operation] returns the current status of [operation]. For a
        copy operation, this will contain progress information.
        """
        return {"failed": True, "complete": True, "progress": None}

    def cancel(self, dbg, operation):
        """
        [cancel operation] cancels a long-running operation. Note that the
        call may return before the operation has finished.
        """

    def destroy(self, dbg, operation):
        """
        [destroy operation] destroys the information about a long-running
        operation. This should fail when run against an operation that is
        still in progress.
        """

    def ls(self, dbg):
        """
        [ls] returns a list of all current operations
        """
        return [None]


class Data_commandline(object):
    """Parse command-line arguments and call an implementation."""

    def __init__(self, impl):
        self.impl = impl
        self.dispatcher = Data_server_dispatcher(self.impl)

    def _parse_copy(self):
        """
        [copy uri domain remotes blocks] copies [blocks] from the local disk
        to a remote URI. This may be called as part of a Volume Mirroring
        operation, and hence may need to cooperate with whatever process is
        currently mirroring writes to ensure data integrity is maintained.
        The [remote] parameter is a remotely accessible URI, for example,
        `nbd://root:pass@foo.com/path/to/disk` that must contain all necessary
        authentication tokens
        """
        # in --json mode we don't have any other arguments
        if ('--json' in sys.argv or '-j' in sys.argv):
            jsondict = json.loads(sys.stdin.readline(),)
            jsondict['json'] = True
            return jsondict
        parser = argparse.ArgumentParser(description='[copy uri domain remotes blocks] copies [blocks] from the local disk  to a remote URI. This may be called as part of a Volume Mirroring  operation, and hence may need to cooperate with whatever process is  currently mirroring writes to ensure data integrity is maintained. The [remote] parameter is a remotely accessible URI, for example, `nbd://root:pass@foo.com/path/to/disk` that must contain all necessary authentication tokens')
        parser.add_argument('-j', '--json', action='store_const', const=True, default=False, help='Read json from stdin, print json to stdout', required=False)
        parser.add_argument('dbg', action='store', help='Debug context from the caller')
        parser.add_argument('uri', action='store', help='A URI which represents how to access the volume disk data.')
        parser.add_argument('domain', action='store', help='An opaque string which represents the Xen domain.')
        parser.add_argument('remote', action='store', help='A URI which represents how to access a remote volume disk data.')
        parser.add_argument('blocklist', action='store', help='List of blocks for copying.')
        return vars(parser.parse_args())

    def _parse_mirror(self):
        """
        [mirror uri domain remote] starts mirroring new writes to the volume
        to a remote URI (usually NBD). This is called as part of a volume
        mirroring process
        """
        # in --json mode we don't have any other arguments
        if ('--json' in sys.argv or '-j' in sys.argv):
            jsondict = json.loads(sys.stdin.readline(),)
            jsondict['json'] = True
            return jsondict
        parser = argparse.ArgumentParser(description='[mirror uri domain remote] starts mirroring new writes to the volume  to a remote URI (usually NBD). This is called as part of a volume  mirroring process')
        parser.add_argument('-j', '--json', action='store_const', const=True, default=False, help='Read json from stdin, print json to stdout', required=False)
        parser.add_argument('dbg', action='store', help='Debug context from the caller')
        parser.add_argument('uri', action='store', help='A URI which represents how to access the volume disk data.')
        parser.add_argument('domain', action='store', help='An opaque string which represents the Xen domain.')
        parser.add_argument('remote', action='store', help='A URI which represents how to access a remote volume disk data.')
        return vars(parser.parse_args())

    def _parse_stat(self):
        """
        [stat operation] returns the current status of [operation]. For a
        copy operation, this will contain progress information.
        """
        # in --json mode we don't have any other arguments
        if ('--json' in sys.argv or '-j' in sys.argv):
            jsondict = json.loads(sys.stdin.readline(),)
            jsondict['json'] = True
            return jsondict
        parser = argparse.ArgumentParser(description='[stat operation] returns the current status of [operation]. For a  copy operation, this will contain progress information.')
        parser.add_argument('-j', '--json', action='store_const', const=True, default=False, help='Read json from stdin, print json to stdout', required=False)
        parser.add_argument('dbg', action='store', help='Debug context from the caller')
        parser.add_argument('operation', action='store', help='The primary key for referring to a long-running operation.')
        return vars(parser.parse_args())

    def _parse_cancel(self):
        """
        [cancel operation] cancels a long-running operation. Note that the
        call may return before the operation has finished.
        """
        # in --json mode we don't have any other arguments
        if ('--json' in sys.argv or '-j' in sys.argv):
            jsondict = json.loads(sys.stdin.readline(),)
            jsondict['json'] = True
            return jsondict
        parser = argparse.ArgumentParser(description='[cancel operation] cancels a long-running operation. Note that the  call may return before the operation has finished.')
        parser.add_argument('-j', '--json', action='store_const', const=True, default=False, help='Read json from stdin, print json to stdout', required=False)
        parser.add_argument('dbg', action='store', help='Debug context from the caller')
        parser.add_argument('operation', action='store', help='The primary key for referring to a long-running operation.')
        return vars(parser.parse_args())

    def _parse_destroy(self):
        """
        [destroy operation] destroys the information about a long-running
        operation. This should fail when run against an operation that is
        still in progress.
        """
        # in --json mode we don't have any other arguments
        if ('--json' in sys.argv or '-j' in sys.argv):
            jsondict = json.loads(sys.stdin.readline(),)
            jsondict['json'] = True
            return jsondict
        parser = argparse.ArgumentParser(description='[destroy operation] destroys the information about a long-running  operation. This should fail when run against an operation that is  still in progress.')
        parser.add_argument('-j', '--json', action='store_const', const=True, default=False, help='Read json from stdin, print json to stdout', required=False)
        parser.add_argument('dbg', action='store', help='Debug context from the caller')
        parser.add_argument('operation', action='store', help='The primary key for referring to a long-running operation.')
        return vars(parser.parse_args())

    def _parse_ls(self):
        """
        [ls] returns a list of all current operations
        """
        # in --json mode we don't have any other arguments
        if ('--json' in sys.argv or '-j' in sys.argv):
            jsondict = json.loads(sys.stdin.readline(),)
            jsondict['json'] = True
            return jsondict
        parser = argparse.ArgumentParser(description='[ls] returns a list of all current operations')
        parser.add_argument('-j', '--json', action='store_const', const=True, default=False, help='Read json from stdin, print json to stdout', required=False)
        parser.add_argument('dbg', action='store', help='Debug context from the caller')
        return vars(parser.parse_args())

    def copy(self):
        """
        [copy uri domain remotes blocks] copies [blocks] from the local disk
        to a remote URI. This may be called as part of a Volume Mirroring
        operation, and hence may need to cooperate with whatever process is
        currently mirroring writes to ensure data integrity is maintained.
        The [remote] parameter is a remotely accessible URI, for example,
        `nbd://root:pass@foo.com/path/to/disk` that must contain all necessary
        authentication tokens
        """
        use_json = False
        try:
            request = self._parse_copy()
            use_json = 'json' in request and request['json']
            results = self.dispatcher.copy(request)
            print(json.dumps(results))
        except Exception as exn:
            if use_json:
                handle_exception(exn)
            else:
                traceback.print_exc()
                raise exn

    def mirror(self):
        """
        [mirror uri domain remote] starts mirroring new writes to the volume
        to a remote URI (usually NBD). This is called as part of a volume
        mirroring process
        """
        use_json = False
        try:
            request = self._parse_mirror()
            use_json = 'json' in request and request['json']
            results = self.dispatcher.mirror(request)
            print(json.dumps(results))
        except Exception as exn:
            if use_json:
                handle_exception(exn)
            else:
                traceback.print_exc()
                raise exn

    def stat(self):
        """
        [stat operation] returns the current status of [operation]. For a
        copy operation, this will contain progress information.
        """
        use_json = False
        try:
            request = self._parse_stat()
            use_json = 'json' in request and request['json']
            results = self.dispatcher.stat(request)
            print(json.dumps(results))
        except Exception as exn:
            if use_json:
                handle_exception(exn)
            else:
                traceback.print_exc()
                raise exn

    def cancel(self):
        """
        [cancel operation] cancels a long-running operation. Note that the
        call may return before the operation has finished.
        """
        use_json = False
        try:
            request = self._parse_cancel()
            use_json = 'json' in request and request['json']
            results = self.dispatcher.cancel(request)
            print(json.dumps(results))
        except Exception as exn:
            if use_json:
                handle_exception(exn)
            else:
                traceback.print_exc()
                raise exn

    def destroy(self):
        """
        [destroy operation] destroys the information about a long-running
        operation. This should fail when run against an operation that is
        still in progress.
        """
        use_json = False
        try:
            request = self._parse_destroy()
            use_json = 'json' in request and request['json']
            results = self.dispatcher.destroy(request)
            print(json.dumps(results))
        except Exception as exn:
            if use_json:
                handle_exception(exn)
            else:
                traceback.print_exc()
                raise exn

    def ls(self):
        """
        [ls] returns a list of all current operations
        """
        use_json = False
        try:
            request = self._parse_ls()
            use_json = 'json' in request and request['json']
            results = self.dispatcher.ls(request)
            print(json.dumps(results))
        except Exception as exn:
            if use_json:
                handle_exception(exn)
            else:
                traceback.print_exc()
                raise exn


class datapath_server_dispatcher(object):
    """Demux calls to individual interface server_dispatchers"""

    def __init__(self, Datapath=None, Data=None):
        self.Datapath = Datapath
        self.Data = Data
        self._dispatcher_dict = {
            "Datapath": self.Datapath._dispatch,
            "Data": self.Data._dispatch
        }

    def _dispatch(self, method, params):
        try:
            logging.log("method = %s params = %s", method, repr(params))
            class_ = method.split('.')[0]
            if class_ in self._dispatcher_dict:
                return self._dispatcher_dict[class_](method, params)
            raise UnknownMethod(method)
        except Rpc_light_failure as rpc_exn:
            logging.log("caught %s", rpc_exn)
            traceback.print_exc()
            logging.log("returning %s", repr(rpc_exn.failure()))
            return rpc_exn.failure()
        except Exception as exn:
            logging.log("caught %s", exn)
            traceback.print_exc()
            return InternalError(get_str(exn)).failure()


class datapath_server_test(datapath_server_dispatcher):
    """Create a server which will respond to all calls, returning arbitrary values. This is intended as a marshal/unmarshal test."""

    def __init__(self):
        datapath_server_dispatcher.__init__(self, Datapath_server_dispatcher(Datapath_test()), Data_server_dispatcher(Data_test()))
