"""Bindings generated for interface plugin by rpclib"""

from __future__ import print_function, division

import argparse
import json
import logging
import sys
import traceback

get_str = str
if sys.version_info[0] > 2:
    long = int
    unicode = str
    str = bytes

from xapi import *


class Plugin_server_dispatcher(object):
    """
    Discover properties of this implementation. Every implementation
    must support the query interface or it will not be recognised as
    a storage plugin by xapi.
    """

    def __init__(self, impl):
        """impl is a proxy object whose methods contain the implementation"""
        self._impl = impl
        self._dispatcher_dict = {
            "Plugin.query": self.query,
            "Plugin.ls": self.ls,
            "Plugin.diagnostics": self.diagnostics
        }

    def query(self, args):
        """type-check inputs, call implementation, type-check outputs and return"""
        if not isinstance(args, dict):
            raise UnmarshalException('arguments', 'dict', repr(args))
        if "dbg" not in args:
            raise UnmarshalException('argument missing', 'dbg', '')
        dbg = args["dbg"]
        if not isinstance(dbg, (str, unicode)):
            raise TypeError("string", repr(dbg))
        results = self._impl.query(dbg)
        if not isinstance(results['plugin'], (str, unicode)):
            raise TypeError("string", repr(results['plugin']))
        if not isinstance(results['name'], (str, unicode)):
            raise TypeError("string", repr(results['name']))
        if not isinstance(results['description'], (str, unicode)):
            raise TypeError("string", repr(results['description']))
        if not isinstance(results['vendor'], (str, unicode)):
            raise TypeError("string", repr(results['vendor']))
        if not isinstance(results['copyright'], (str, unicode)):
            raise TypeError("string", repr(results['copyright']))
        if not isinstance(results['version'], (str, unicode)):
            raise TypeError("string", repr(results['version']))
        if not isinstance(results['required_api_version'], (str, unicode)):
            raise TypeError("string", repr(results['required_api_version']))
        if not isinstance(results['features'], list):
            raise TypeError("string list", repr(results['features']))
        for tmp_9 in results['features']:
            if not isinstance(tmp_9, (str, unicode)):
                raise TypeError("string", repr(tmp_9))
        if not isinstance(results['configuration'], dict):
            raise TypeError("(string * string) list", repr(results['configuration']))
        for tmp_7, tmp_8 in results['configuration'].items():
            if not isinstance(tmp_7, (str, unicode)):
                raise TypeError("string", repr(tmp_7))
            if not isinstance(tmp_8, (str, unicode)):
                raise TypeError("string", repr(tmp_8))
        if not isinstance(results['required_cluster_stack'], list):
            raise TypeError("string list", repr(results['required_cluster_stack']))
        for tmp_6 in results['required_cluster_stack']:
            if not isinstance(tmp_6, (str, unicode)):
                raise TypeError("string", repr(tmp_6))
        return results

    def ls(self, args):
        """type-check inputs, call implementation, type-check outputs and return"""
        if not isinstance(args, dict):
            raise UnmarshalException('arguments', 'dict', repr(args))
        if "dbg" not in args:
            raise UnmarshalException('argument missing', 'dbg', '')
        dbg = args["dbg"]
        if not isinstance(dbg, (str, unicode)):
            raise TypeError("string", repr(dbg))
        results = self._impl.ls(dbg)
        if not isinstance(results, list):
            raise TypeError("string list", repr(results))
        for tmp_10 in results:
            if not isinstance(tmp_10, (str, unicode)):
                raise TypeError("string", repr(tmp_10))
        return results

    def diagnostics(self, args):
        """type-check inputs, call implementation, type-check outputs and return"""
        if not isinstance(args, dict):
            raise UnmarshalException('arguments', 'dict', repr(args))
        if "dbg" not in args:
            raise UnmarshalException('argument missing', 'dbg', '')
        dbg = args["dbg"]
        if not isinstance(dbg, (str, unicode)):
            raise TypeError("string", repr(dbg))
        results = self._impl.diagnostics(dbg)
        if not isinstance(results, (str, unicode)):
            raise TypeError("string", repr(results))
        return results

    def _dispatch(self, method, params):
        """type check inputs, call implementation, type check outputs and return"""
        args = params[0]
        return success(self._dispatcher_dict[method](args))


class Plugin_skeleton(object):
    """
    Discover properties of this implementation. Every implementation
    must support the query interface or it will not be recognised as
    a storage plugin by xapi.
    """

    def __init__(self):
        pass

    def query(self, dbg):
        """
        Query this implementation and return its properties. This is
        called by xapi to determine whether it is compatible with xapi
        and to discover the supported features.
        """
        raise Unimplemented("Plugin.query")

    def ls(self, dbg):
        """
        [ls dbg]: returns a list of attached SRs
        """
        raise Unimplemented("Plugin.ls")

    def diagnostics(self, dbg):
        """
        Returns a printable set of backend diagnostic information.
        Implementations are encouraged to include any data which will
        be useful to diagnose problems. Note this data should not
        include personally-identifiable data as it is intended to be
        automatically included in bug reports.
        """
        raise Unimplemented("Plugin.diagnostics")


class Plugin_test(object):
    """
    Discover properties of this implementation. Every implementation
    must support the query interface or it will not be recognised as
    a storage plugin by xapi.
    """

    def __init__(self):
        pass

    def query(self, dbg):
        """
        Query this implementation and return its properties. This is
        called by xapi to determine whether it is compatible with xapi
        and to discover the supported features.
        """
        return {"plugin": "string", "name": "string", "description": "string", "vendor": "string", "copyright": "string", "version": "string", "required_api_version": "string", "features": ["string"], "configuration": {"string": "string"}, "required_cluster_stack": ["string"]}

    def ls(self, dbg):
        """
        [ls dbg]: returns a list of attached SRs
        """
        return ["string"]

    def diagnostics(self, dbg):
        """
        Returns a printable set of backend diagnostic information.
        Implementations are encouraged to include any data which will
        be useful to diagnose problems. Note this data should not
        include personally-identifiable data as it is intended to be
        automatically included in bug reports.
        """
        return "string"


class Plugin_commandline(object):
    """Parse command-line arguments and call an implementation."""

    def __init__(self, impl):
        self.impl = impl
        self.dispatcher = Plugin_server_dispatcher(self.impl)

    def _parse_query(self):
        """
        Query this implementation and return its properties. This is
        called by xapi to determine whether it is compatible with xapi
        and to discover the supported features.
        """
        # in --json mode we don't have any other arguments
        if ('--json' in sys.argv or '-j' in sys.argv):
            jsondict = json.loads(sys.stdin.readline(),)
            jsondict['json'] = True
            return jsondict
        parser = argparse.ArgumentParser(description='Query this implementation and return its properties. This is  called by xapi to determine whether it is compatible with xapi  and to discover the supported features.')
        parser.add_argument('-j', '--json', action='store_const', const=True, default=False, help='Read json from stdin, print json to stdout', required=False)
        parser.add_argument('dbg', action='store', help='Debug context from the caller')
        return vars(parser.parse_args())

    def _parse_ls(self):
        """
        [ls dbg]: returns a list of attached SRs
        """
        # in --json mode we don't have any other arguments
        if ('--json' in sys.argv or '-j' in sys.argv):
            jsondict = json.loads(sys.stdin.readline(),)
            jsondict['json'] = True
            return jsondict
        parser = argparse.ArgumentParser(description='[ls dbg]: returns a list of attached SRs')
        parser.add_argument('-j', '--json', action='store_const', const=True, default=False, help='Read json from stdin, print json to stdout', required=False)
        parser.add_argument('dbg', action='store', help='Debug context from the caller')
        return vars(parser.parse_args())

    def _parse_diagnostics(self):
        """
        Returns a printable set of backend diagnostic information.
        Implementations are encouraged to include any data which will
        be useful to diagnose problems. Note this data should not
        include personally-identifiable data as it is intended to be
        automatically included in bug reports.
        """
        # in --json mode we don't have any other arguments
        if ('--json' in sys.argv or '-j' in sys.argv):
            jsondict = json.loads(sys.stdin.readline(),)
            jsondict['json'] = True
            return jsondict
        parser = argparse.ArgumentParser(description='Returns a printable set of backend diagnostic information.  Implementations are encouraged to include any data which will  be useful to diagnose problems. Note this data should not  include personally-identifiable data as it is intended to be   automatically included in bug reports.')
        parser.add_argument('-j', '--json', action='store_const', const=True, default=False, help='Read json from stdin, print json to stdout', required=False)
        parser.add_argument('dbg', action='store', help='Debug context from the caller')
        return vars(parser.parse_args())

    def query(self):
        """
        Query this implementation and return its properties. This is
        called by xapi to determine whether it is compatible with xapi
        and to discover the supported features.
        """
        use_json = False
        try:
            request = self._parse_query()
            use_json = 'json' in request and request['json']
            results = self.dispatcher.query(request)
            print(json.dumps(results))
        except Exception as exn:
            if use_json:
                handle_exception(exn)
            else:
                traceback.print_exc()
                raise exn

    def ls(self):
        """
        [ls dbg]: returns a list of attached SRs
        """
        use_json = False
        try:
            request = self._parse_ls()
            use_json = 'json' in request and request['json']
            results = self.dispatcher.ls(request)
            print(json.dumps(results))
        except Exception as exn:
            if use_json:
                handle_exception(exn)
            else:
                traceback.print_exc()
                raise exn

    def diagnostics(self):
        """
        Returns a printable set of backend diagnostic information.
        Implementations are encouraged to include any data which will
        be useful to diagnose problems. Note this data should not
        include personally-identifiable data as it is intended to be
        automatically included in bug reports.
        """
        use_json = False
        try:
            request = self._parse_diagnostics()
            use_json = 'json' in request and request['json']
            results = self.dispatcher.diagnostics(request)
            print(json.dumps(results))
        except Exception as exn:
            if use_json:
                handle_exception(exn)
            else:
                traceback.print_exc()
                raise exn


class plugin_server_dispatcher(object):
    """Demux calls to individual interface server_dispatchers"""

    def __init__(self, Plugin=None):
        self.Plugin = Plugin
        self._dispatcher_dict = {
            "Plugin": self.Plugin._dispatch
        }

    def _dispatch(self, method, params):
        try:
            logging.log("method = %s params = %s", method, repr(params))
            class_ = method.split('.')[0]
            if class_ in self._dispatcher_dict:
                return self._dispatcher_dict[class_](method, params)
            raise UnknownMethod(method)
        except Rpc_light_failure as rpc_exn:
            logging.log("caught %s", rpc_exn)
            traceback.print_exc()
            logging.log("returning %s", repr(rpc_exn.failure()))
            return rpc_exn.failure()
        except Exception as exn:
            logging.log("caught %s", exn)
            traceback.print_exc()
            return InternalError(get_str(exn)).failure()


class plugin_server_test(plugin_server_dispatcher):
    """Create a server which will respond to all calls, returning arbitrary values. This is intended as a marshal/unmarshal test."""

    def __init__(self):
        plugin_server_dispatcher.__init__(self, Plugin_server_dispatcher(Plugin_test()))
